"""Module to convert phonemes."""


def multi_character_tokenize(line, multi_character_tokens):
	"""Tries to match one of the tokens in multi_character_tokens at each position of line, starting at position 0,
	if so tokenizes and eats that token. Otherwise tokenizes a single character"""
	while line != '':
		for token in multi_character_tokens:
			if line.startswith(token) and len(token) > 0:
				yield token
				line = line[len(token):]
				break
		else:
			yield line[:1]
			line = line[1:]


def split_ipa(line):
	"""
	Split a line by IPA phones.
	If nasalized sound (such as ɛ̃ː) is included, it will give error.
	:param string line: one line written in IPA.
	:return string lineSeperated: the line splitted in IPA phone. 
	"""

	multi_character_phonemes = [
		# IPAs in CGN.
		u'ʌu', u'ɛi', u'œy', u'aː', u'eː', u'iː', u'oː', u'øː', u'ɛː', u'œː', u'ɔː', u'ɛ̃ː', u'ɑ̃ː', u'ɔ̃ː', u'œ̃', u'ɪː'
		]

	return ['sp' if phoneme == ' ' else phoneme
			for phoneme in multi_character_tokenize(line.strip(), multi_character_phonemes)]


def split_fame_ipa(line):
	"""
	Split a line by IPA phones.
	If nasalized sound (such as ɛ̃ː) is included, it will give error.
	:param string line: one line written in IPA used in FAME! corpus.
	:return string lineSeperated: the line splitted in IPA phone. 
	"""

	multi_character_phonemes = [
		# IPAs in the Fame! corpus.
		u'a:', u's:', u'r:', u'i:', u'ɪ:', u'i̯', u'e:', u'ə:', u'ɛ:', u'u:', u'ü:', u'ü', u'ṷ', u'ö:', u'o:', u'ɔ:', u'ɔ̈:', u'ɔ̈',
        # IPAs in stimmen.
        u'ø:', u'ɑ:', u'ɐ:', u'ɒ:', u'œ:', u'y:'
		]

	return ['sp' if phoneme == ' ' else phoneme
			for phoneme in multi_character_tokenize(line.strip(), multi_character_phonemes)]

		
def split_cgn(line):
	""" Split a line by CGN phones.
	
	:param string line: one line written in CGN phonesets.
	:return string lineSeperated: the line splitted by each CGN phone. 
	"""

	multi_character_phonemes = ['A+', 'A~', 'E+', 'E:', 'E~', 'O:', 'O~', 'U~', 'Y+', 'Y:']

	return ['sp' if phoneme == ' ' else
			('C' if phoneme == '2' else phoneme)
			for phoneme in  multi_character_tokenize(line.strip(), multi_character_phonemes)
			if phoneme != '_'
			]


def ipa2cgn(tokens):
	"""Convert IPA symbols to CGN phonemes.
	
	:param string line: one line written in IPA.
	:return string lineConverted: one line written in CGN phonesets. 
	"""
	translation_key = {'ə': '@', 'ɑ': 'A', 'ʌu': 'A+', 'ɛ': 'E', 'ɛi': 'E+', 'øː': 'C', 'ɣ': 'G', 'ɪ': 'I', 'ŋ': 'N',
					   'ɔ': 'O', 'ʃ': 'S', 'ʏ': 'Y', 'œy': 'Y+', 'ʒ': 'Z', 'aː': 'a', 'eː': 'e', 'ɦ': 'h', 'iː': 'i',
					   'oː': 'o', 'ʋ': 'w', 'ɛː': 'E:', 'ɦ': 'h', 'iː': 'i', 'œː': 'Y:', 'ɔː': 'O:', 'ɛ̃ː': 'E~',
					   'ɑ̃ː': 'A~', 'ɔ̃ː': 'O~', 'œ̃': 'U~',
					   # Aki added
					   'ʀ': 'r', 'ɪː': 'i', 'ʉ': 'u', 'ː': ':'}
	return ' '.join([
		''.join(
			translation_key.get(phoneme, phoneme)
			for phoneme in split_ipa(word)
		)
		for word in tokens
	])
		
		
def cgn2barbara(tokens):
	"""
	Convert CGN phoneset to Barbara's phoneset.
	:param string line: one line written in CGN phoneset.
	:return string: one line written in Barbara's phonesets. 
	"""
	# replace phonemes in CGN with phonemes of Barbara
	translation_key = {'C': 'eu', 'A': 'ac', 'A+': 'au', 'A~': 'a', 'E': 'ec', 'E+': 'ei', 'E~': 'ec', 'E:': 'ec', 'G': 'gc',
			   'I': 'ic', 'N': 'nc', 'O': 'oc', 'O:': 'oc', 'O~': 'oc', 'silB': 'ssil', 'silE': 'ssil', 'sp': 'ssil',
			   'S': 'sc', 'Y': 'yc', 'Y+': 'ui', 'Y:': 'yc', 'Z': 'z', 'U~': 'yc',
			   # As per Aki's suggestion by e-mail on 2018-01-05:
			   '|': 'eu', 'K': 'ei',
			   # Aki suggests on 2018-03-17:
			   '}': 'u', 'L': 'ui'
					   }
	return ' '.join([
		''.join(
			translation_key.get(phoneme, phoneme)
			for phoneme in split_cgn(word)
		)
		for word in tokens
	])


def ipa2famehtk(tokens):
	"""
	Convert IPA phoneset to the phoneset for FAME! acoustic model in HTK.
	:param string line: one line written in IPA phoneset.
	:return string: one line written in the phoneset for FAME! acoustic model in HTK. 
	"""

	translation_key = {'ŋ': 'ng', 's:': 'sh', 
					   #'r:': 'rh', 
					   'ɡ': 'q',
					   'i̯': 'iu', 'i:': 'ih', 'ɪ': 'ic', 'ɪ:': 'ich', 
					   'e:': 'eh', 
					   'ə': 'ec',  
					   #'ə:': 'ech', 
					   'ɛ': 'w', 'ɛ:': 'wh',
					   'a:': 'ah',
					   'ṷ': 'uv', 'ú': 'ua', 
					   'u:': 'uh', 
					   'ü': 'ue', 'ü:': 'ueh',
					   'o:': 'oh',
					   'ö': 'oe', 'ö:': 'oeh',
					   'ɔ': 'c', 'ɔ:': 'ch',
					   'ɔ̈': 'ce', u'ɔ̈:': 'ceh',
					   # based on the recommendation from Martijn on 2018/04/09.
					   'y': 'ih', 'e': 'eh', 'ə:': 'wh', 'r:': 'r',
					   
                       # ipa which is included in stimmem, but not in fame!
					   # based on the recommendation from Jelske on 2018/06/21.
					   'æ': 'w', 
                       'ø': 'oe', 
                       'œ': 'ce', 
                       'ɐ': 'a', 'ɑ': 'a', 
					   'ɒ': 'ce', 
                       'ɾ': 'r', 'ʁ': 'r', 
                       'ʊ': 'u', 
                       'χ': 'x',
                       # after bugfix regarding '\\'
                       'ø:': 'oeh', 
                       'œ:': 'wh', 
                       'ɐ:': 'ah', 'ɑ:': 'ah',
                       'ɒ:': 'ce', # should be ceh, but ceh is ignored. 
                       'y:': 'ih', 
                       # ???
                       'ɹ': 'v', 'ʀ': 'r'
					   }
	#return ' '.join([
	#	''.join(
	#		translation_key.get(phoneme, phoneme)
	#		for phoneme in split_ipa(tokens)
	#	)
	#	for word in tokens
	#])
	pronunciation = []
	for phoneme in split_fame_ipa(tokens):
		pronunciation.append(translation_key.get(phoneme, phoneme))
	return ' '.join(pronunciation)


def ipa2cgnFile(fileIPA, fileCGN):
	"""
	Convert an IPA file into a CGN file.
	Both files should be written like [word]/t[pronunciation] in every line. 
	:param path fileIPA: path to the file in which pronunciations are written in IPA.
	:param path fileCGN: path to the file in which pronunciations are written in CGN phoneset. 
	"""
	with open(fileCGN, 'w', encoding='utf-8') as fout:
		with open(fileIPA, 'r', encoding='utf-8') as fin:
			for line in fin:
				word, ipa = line.strip().split()[:2]
				cgn = ipa2cgn(ipa.strip().split(' '))
				fout.write("{0}\t{1}\n".format(word, cgn))


def ipa2barbara_htk(line):
	"""
	Convert IPA to Barbara's description in HTK format as follows:
	- start and end with 'ssil'. 
	- each phones are written in Barbara's description.
	- each phones are splitted with a space.
	:param string line: a line written in IPA.
	:returns string barbara_htk: a line written in Barbara's phoneset in HTK format.
	"""
	cgn = ipa2cgn(line)
	cgn_split = split_cgn(cgn)
	barbara_htk = 'ssil ' + cgn2barbara(cgn_split) + ' ssil'
	return barbara_htk